/*
 * MODULE
 *
 *   init.c
 *
 * DESCRIPTION
 *
 *  Hardware Initialization Routines
 *
 * PLATFORM
 *
 *  Olimex LPC-2478-STK
 *
 * HISTORY
 *
 *   2013/02/10: John Hatch:
 *   	Rewrote PinConfigurator for new header
 *   2012/11/21: John Hatch:
 *   	Converted to use lpc2478.h header
 *   2007/10/31: Mitch Ishihara:
 *     changed PINSEL10 setting to just write the value instead
 *       of read-modify-write.
 *   2007/10/01: Mitch Ishihara:
 *     split init routines from main.c
 *     changed MEMMAP to USERRAMMODE to run completely from RAM
 *   2007/09/12: Andrew N. Sloss:
 *     convert the main.c to use VOLATILExx(<address>) format
 *     added 1st draft comments
 *   2007/08/21: Andrew N. Sloss:
 *     inspired from James Lynch's example code demo2106_blink_flash.
 *
 * NOTES
 *
 *  Taken from the LPC2378 User Manual:
 *
 *   5.14 PLL setup sequence
 *
 *   The following sequence must be followed step by step in order to
 *   have the PLL initialized an running:
 *
 *   1. Disconnect the PLL with one feed sequence if PLL is already
 *      connected.
 *   2. Disable the PLL with one feed sequence.
 *   3. Change the CPU Clock Divider setting to speed up operation without
 *      the PLL, if desired.
 *   4. Write to the Clock Source Selection Control register to change the
 *      clock source.
 *   5. Write to the PLLCFG and make it effective with one feed sequence.
 *      The PLLCFG can only be updated when the PLL is disabled.
 *   6. Enable the PLL with one feed sequence.
 *   7. Change the CPU Clock Divider setting for the operation with the
 *      PLL. It's critical to do this before connecting the PLL.
 *   8. Wait for the PLL to achieve lock by monitoring the PLOCK bit in the
 *      PLLSTAT register, or using the PLOCK interrupt, or wait for a fixed
 *      time when the input clock to PLL is slow (i.e. 32 kHz). The value of
 *      PLOCK may not be stable when the PLL reference frequency (FREF, the
 *      frequency of REFCLK, which is equal to the PLL input frequency divided
 *      by the pre-divider value) is less than 100 kHz or greater than 20
 *      MHz. In these cases, the PLL may be assumed to be stable after a
 *      start-up time has passed. This time is 500 us when FREF is
 *      greater than 400 kHz and 200 / FREF seconds when FREF is less than
 *      400kHz.
 *   9. Connect the PLL with one feed sequence.
 *
 *   It's very important not to merge any steps above. For example, don't
 *   update the PLLCFG and enable the PLL simultaneously with the same feed
 *   sequence.
 *
 */

#include "lpc2478.h"
#include "init.h"

// Macros for Pin configuration
#define PRIMARY            0x0              /* 0b00 primary function */
#define FIRST              0x1              /* 0b01 first  alternate function */
#define SECOND             0x2              /* 0b10 second alternate function */
#define THIRD              0x3              /* 0b11 third  alternate function */

//
//  InitPLL
//
//  This init is specific to the 2300 family of devices.
//
//  Initialize PLL (Configured for a 12.000 MHz crystal) to
//  boost processor clock to 72.000 MHz
//
static void InitPLL(void)
{
    // [1] Check if PLL connected, disconnect if yes.

    if (PLLSTAT & PLLSTAT_PLLC)
    {
        PLLCON = PLLCON_PLLE;
        // Enable PLL, disconnected ( PLLC = 0)
        PLLFEED = 0xaa;
        PLLFEED = 0x55;
    }

    // [2] Disable the PLL once it has been disconnected.

    PLLCON  = 0;
    PLLFEED = 0xaa;
    PLLFEED = 0x55;

    /* Enable the main oscillator */

    SCS |= SCS_OSCEN;

    // Wait until main OSC is usable

    do {} while ((SCS & SCS_OSCSTAT) == 0);

    // Select the main osc as the PLL clk source.

    CLKSRCSEL = CLKSRC_MAIN_OSC;

	// Set multiplier and divider values.
	//
	//  PLLCFG = ((N - 1) << 16) + (M - 1)
	//
	//  F_cco = (2 * M *F_in)/N
	//      M = 12, N = 1, F_in = 12.000 MHz -> F_cco = 288.000 MHz
    
    PLLCFG = PLLCFG_VALUE;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;
    
    // Enable the PLL.
    
    PLLCON  = PLLCON_PLLE;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;
    
    // Divide F_cco down to get the CCLK output. (288 / 6 = 48)
    
    CCLKCFG = CCLKCFG_VALUE;
    
    // Wait for the PLL to lock to set frequency
    
    do {} while((PLLSTAT & PLLSTAT_PLOCK) == 0);
    
    do {} while((PLLSTAT & 0x00FF7FFF) != PLLCFG_VALUE);
    
    // Enable and connect the PLL as the clock source
    
    PLLCON = (PLLCON_PLLE | PLLCON_PLLC);
    PLLFEED = 0xAA;
    PLLFEED = 0x55;
    
    // Check connect bit status and wait for connection.
    
    do {} while((PLLSTAT & PLLSTAT_PLLC) == 0);
}

//
//  InitMAM
//
//  Eenable the MAM accelerator and sets number of clocks used for
//  flash memory fetch.
//
static void InitMAM(void)
{
    MAMCR = MAMCR_OFF;
    MAMTIM = MAMTIM_3_CLK;
    MAMCR = MAMCR_PART;
}

//
//  InitPCLK
//
//  Set the pclk for each peripheral. The default PDIV_4, but we do that
//  explicitly here to make it easier to change them later.
//
#define PDIV_1             0x1
#define PDIV_2             0x2
#define PDIV_4             0x0
#define PDIV_6             0x3
#define PDIV_8             0x3

#define SET_PCLK(per, div) ((div) << (per))

static void InitPCLK(void)
{
    PCLKSEL0_bit.PCLK_WDT    = PDIV_4;
    PCLKSEL0_bit.PCLK_TIMER0 = PDIV_4;
    PCLKSEL0_bit.PCLK_TIMER1 = PDIV_4;
    PCLKSEL0_bit.PCLK_UART0  = PDIV_1;
    PCLKSEL0_bit.PCLK_UART1  = PDIV_1;
    PCLKSEL0_bit.PCLK_PWM0   = PDIV_4;
    PCLKSEL0_bit.PCLK_PWM1   = PDIV_4;
    PCLKSEL0_bit.PCLK_I2C0   = PDIV_4;
    PCLKSEL0_bit.PCLK_SPI    = PDIV_4;
    PCLKSEL0_bit.PCLK_RTC    = PDIV_4;
    PCLKSEL0_bit.PCLK_SSP1   = PDIV_4;
    PCLKSEL0_bit.PCLK_DAC    = PDIV_4;
    PCLKSEL0_bit.PCLK_ADC    = PDIV_4;
    PCLKSEL0_bit.PCLK_CAN1   = PDIV_4;
    PCLKSEL0_bit.PCLK_CAN2   = PDIV_4;
    PCLKSEL0_bit.PCLK_ACF    = PDIV_4;

    PCLKSEL1_bit.PCLK_BAT_RAM = PDIV_4;
    PCLKSEL1_bit.PCLK_GPIO    = PDIV_4;
    PCLKSEL1_bit.PCLK_PCB     = PDIV_4;
    PCLKSEL1_bit.PCLK_I2C1    = PDIV_4;
    PCLKSEL1_bit.PCLK_SSP0    = PDIV_4;
    PCLKSEL1_bit.PCLK_TIMER2  = PDIV_4;
    PCLKSEL1_bit.PCLK_TIMER3  = PDIV_4;
    PCLKSEL1_bit.PCLK_UART2   = PDIV_1;
    PCLKSEL1_bit.PCLK_UART3   = PDIV_1;
    PCLKSEL1_bit.PCLK_I2C2    = PDIV_4;
    PCLKSEL1_bit.PCLK_I2S     = PDIV_4;
    PCLKSEL1_bit.PCLK_MCI     = PDIV_4;
    PCLKSEL1_bit.PCLK_SYSCON  = PDIV_4;
}

//
//  pinConfigurator
//
//  Setups all the pins of the LPC2478
//
void PinConfigurator(void)
{
	PINSEL0_bit.P0_0 = PRIMARY;
	PINSEL0_bit.P0_1 = PRIMARY;
	PINSEL0_bit.P0_2 = FIRST;
	PINSEL0_bit.P0_3 = FIRST;
	PINSEL0_bit.P0_4 = PRIMARY;
	PINSEL0_bit.P0_5 = PRIMARY;
	PINSEL0_bit.P0_6 = PRIMARY;
	PINSEL0_bit.P0_7 = PRIMARY;
	PINSEL0_bit.P0_8 = PRIMARY;
	PINSEL0_bit.P0_9 = PRIMARY;
	PINSEL0_bit.P0_10 = PRIMARY;
	PINSEL0_bit.P0_11 = PRIMARY;
	PINSEL0_bit.P0_12 = PRIMARY;
	PINSEL0_bit.P0_13 = PRIMARY;
	PINSEL0_bit.P0_14 = PRIMARY;
	PINSEL0_bit.P0_15 = PRIMARY;

	PINSEL1_bit.P0_16 = PRIMARY;
	PINSEL1_bit.P0_17 = PRIMARY;
	PINSEL1_bit.P0_18 = PRIMARY;
	PINSEL1_bit.P0_19 = PRIMARY;
	PINSEL1_bit.P0_20 = PRIMARY;
	PINSEL1_bit.P0_21 = PRIMARY;
	PINSEL1_bit.P0_22 = PRIMARY;
	PINSEL1_bit.P0_23 = PRIMARY;
	PINSEL1_bit.P0_24 = PRIMARY;
	PINSEL1_bit.P0_25 = PRIMARY;
	PINSEL1_bit.P0_26 = PRIMARY;
	PINSEL1_bit.P0_27 = PRIMARY;
	PINSEL1_bit.P0_28 = PRIMARY;
	PINSEL1_bit.P0_29 = PRIMARY;
	PINSEL1_bit.P0_30 = PRIMARY;
	PINSEL1_bit.P0_31 = PRIMARY;

	PINSEL2_bit.P1_0 = PRIMARY;
	PINSEL2_bit.P1_1 = PRIMARY;
	PINSEL2_bit.P1_2 = PRIMARY;
	PINSEL2_bit.P1_3 = PRIMARY;
	PINSEL2_bit.P1_4 = PRIMARY;
	PINSEL2_bit.P1_5 = PRIMARY;  //11:10 P1[5]:  GPIO Port 1.5 | ENET_TX_ER | MCIPWR | PWM0[3]
	PINSEL2_bit.P1_6 = PRIMARY;
	PINSEL2_bit.P1_7 = PRIMARY;
	PINSEL2_bit.P1_8 = PRIMARY;
	PINSEL2_bit.P1_9 = PRIMARY;
	PINSEL2_bit.P1_10 = PRIMARY;
	PINSEL2_bit.P1_11 = PRIMARY;
	PINSEL2_bit.P1_12 = PRIMARY;
	PINSEL2_bit.P1_13 = PRIMARY;
	PINSEL2_bit.P1_14 = PRIMARY;
	PINSEL2_bit.P1_15 = PRIMARY;

	PINSEL3_bit.P1_16 = PRIMARY;
	PINSEL3_bit.P1_17 = PRIMARY;
	PINSEL3_bit.P1_18 = PRIMARY;
	PINSEL3_bit.P1_19 = PRIMARY;
	PINSEL3_bit.P1_20 = PRIMARY;
	PINSEL3_bit.P1_21 = PRIMARY;
	PINSEL3_bit.P1_22 = PRIMARY;
	PINSEL3_bit.P1_23 = PRIMARY;
	PINSEL3_bit.P1_24 = PRIMARY;
	PINSEL3_bit.P1_25 = PRIMARY;
	PINSEL3_bit.P1_26 = PRIMARY;
	PINSEL3_bit.P1_27 = PRIMARY;
	PINSEL3_bit.P1_28 = PRIMARY;
	PINSEL3_bit.P1_29 = PRIMARY;
	PINSEL3_bit.P1_30 = PRIMARY;
	PINSEL3_bit.P1_31 = PRIMARY;

	PINSEL4_bit.P2_0 = PRIMARY;
	PINSEL4_bit.P2_1 = PRIMARY;
	PINSEL4_bit.P2_2 = PRIMARY;
	PINSEL4_bit.P2_3 = PRIMARY;
	PINSEL4_bit.P2_4 = PRIMARY;
	PINSEL4_bit.P2_5 = PRIMARY;
	PINSEL4_bit.P2_6 = PRIMARY;
	PINSEL4_bit.P2_7 = PRIMARY;
	PINSEL4_bit.P2_8 = PRIMARY;
	PINSEL4_bit.P2_9 = PRIMARY;
	PINSEL4_bit.P2_10 = PRIMARY;
	PINSEL4_bit.P2_11 = PRIMARY;
	PINSEL4_bit.P2_12 = PRIMARY;
	PINSEL4_bit.P2_13 = PRIMARY;
	PINSEL4_bit.P2_14 = PRIMARY;
	PINSEL4_bit.P2_15 = PRIMARY;

	PINSEL5_bit.P2_16 = PRIMARY;
	PINSEL5_bit.P2_17 = PRIMARY;
	PINSEL5_bit.P2_18 = PRIMARY;
	PINSEL5_bit.P2_19 = PRIMARY;
	PINSEL5_bit.P2_20 = PRIMARY;
	PINSEL5_bit.P2_21 = PRIMARY;
	PINSEL5_bit.P2_22 = PRIMARY;
	PINSEL5_bit.P2_23 = PRIMARY;
	PINSEL5_bit.P2_24 = PRIMARY;
	PINSEL5_bit.P2_25 = PRIMARY;
	PINSEL5_bit.P2_26 = PRIMARY;
	PINSEL5_bit.P2_27 = PRIMARY;
	PINSEL5_bit.P2_28 = PRIMARY;
	PINSEL5_bit.P2_29 = PRIMARY;
	PINSEL5_bit.P2_30 = PRIMARY;
	PINSEL5_bit.P2_31 = PRIMARY;

	PINSEL6_bit.P3_0 = PRIMARY;
	PINSEL6_bit.P3_1 = PRIMARY;
	PINSEL6_bit.P3_2 = PRIMARY;
	PINSEL6_bit.P3_3 = PRIMARY;
	PINSEL6_bit.P3_4 = PRIMARY;
	PINSEL6_bit.P3_5 = PRIMARY;
	PINSEL6_bit.P3_6 = PRIMARY;
	PINSEL6_bit.P3_7 = PRIMARY;
	PINSEL6_bit.P3_8 = PRIMARY;
	PINSEL6_bit.P3_9 = PRIMARY;
	PINSEL6_bit.P3_10 = PRIMARY;
	PINSEL6_bit.P3_11 = PRIMARY;
	PINSEL6_bit.P3_12 = PRIMARY;
	PINSEL6_bit.P3_13 = PRIMARY;
	PINSEL6_bit.P3_14 = PRIMARY;
	PINSEL6_bit.P3_15 = PRIMARY;

	PINSEL7_bit.P3_16 = PRIMARY;
	PINSEL7_bit.P3_17 = PRIMARY;
	PINSEL7_bit.P3_18 = PRIMARY;
	PINSEL7_bit.P3_19 = PRIMARY;
	PINSEL7_bit.P3_20 = PRIMARY;
	PINSEL7_bit.P3_21 = PRIMARY;
	PINSEL7_bit.P3_22 = PRIMARY;
	PINSEL7_bit.P3_23 = PRIMARY;
	PINSEL7_bit.P3_24 = PRIMARY;
	PINSEL7_bit.P3_25 = PRIMARY;
	PINSEL7_bit.P3_26 = PRIMARY;
	PINSEL7_bit.P3_27 = PRIMARY;
	PINSEL7_bit.P3_28 = PRIMARY;
	PINSEL7_bit.P3_29 = PRIMARY;
	PINSEL7_bit.P3_30 = PRIMARY;
	PINSEL7_bit.P3_31 = PRIMARY;

	// Note: Complete by setting the ETM to be disabled on PINSEL10

    // Disable ETM, mask reserved bit to 0
    PINSEL10 = ~(BIT(3)) & PINSEL10_MASK;
}

//
//  InitGPIO
//
//  Sets up the GPIO pins
//
static void InitGPIO(void)
{
    PinConfigurator();

	// NOTE:
	//
	//  SCS controls whether the LPC2378 is set to use the legancy registers or
	//  the new fast GPIO control registers.

    // Fast GPIO / new registers
    SCS |= BIT(0);
}

//
//  initUART0
//
//  Sets up UART0 to be 38.4 K Baud, 8 data bits, no parity, and no flow control
//
void InitUART0(uint16_t baud, uint8_t mode, uint8_t fmode)
{
    U0LCR = ULCR_DLAB_ENABLE;
    U0DLL = (uint8_t) baud;
    U0DLM = (uint8_t)(baud >> 8);
    U0LCR = (mode & ~ULCR_DLAB_ENABLE);
    U0FCR = fmode;
}

//
//   InitHardware
//
//   Initializes all the hardware components for the UART0 to function.
//
void InitHardware(void)
{
    // intialize main hardware components
    
    InitPLL();
    InitMAM();
    InitPCLK();
    InitGPIO();
    
    // intialize specific hardware components for UART0
    InitUART0(UART_BAUD(HOST_BAUD_U0), UART_8N1, UART_FIFO_OFF);

    // Turn off MCIPWR SD LED (near bottom left corner of LCD)
    FIO1DIR_bit.P1_5 = 1;
    FIO1CLR_bit.P1_5 = 1;
    
    // MEMMAP Choices are:
    // BOOTLOADERMODE      0x00
    // USERFLASHMODE       0x01
    // USERRAMMODE         0x02
    // EXTERNALMEMORYMODE  0x03

    MEMMAP = USERFLASHMODE;
}


